/*
Copyright (c) 2001 The University of Texas
All Rights Reserved.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

The GNU Public License is available in the file LICENSE, or you
can write to the Free Software Foundation, Inc., 59 Temple Place -
Suite 330, Boston, MA 02111-1307, USA, or you can find it on the
World Wide Web at http://www.fsf.org.

Programmers:    Wade Schwartzkopf
Version:        @(#)classify.c    1.0     08/16/02

The author performed this work while he was with the Department of
Electrical and Computer Engineering, The University of Texas at Austin,
Austin, TX.  Wade Schwartzkopf can be reached at wade@schwartzkopf.org.
The professors supervising the work were Prof. Alan C. Bovik,
bovik@ece.utexas.edu, and Prof. Brian L. Evans, bevans@ece.utexas.edu.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "classify.h"

#define FILESEP '/'

// Allocate memory for training data data structure
int initTrainingDataStruct(struct trainDataStruct* trainData, int classes, int spectra)
{
   int i,j;

   trainData->classes=classes;
   trainData->spectra=spectra;

   // Allocate memory for means
   if(!(trainData->means=(float **) malloc(classes*sizeof(float*)))) {
      return 0;
   }
   for(i=0;i<classes;i++) {
      if(!(trainData->means[i]=(float *) malloc(spectra*sizeof(float)))) {
         return 0;
      }
   }
   // Allocate memory for inverse covariance matrices
   if(!(trainData->covarianceInverses=(float ***) malloc(classes*sizeof(float**)))) {
      return 0;
   }
   for(i=0;i<classes;i++) {
      if(!(trainData->covarianceInverses[i]=(float **) malloc(spectra*sizeof(float*)))) {
         return 0;
      }
      for(j=0;j<spectra;j++) {
	 if(!(trainData->covarianceInverses[i][j]=(float *) malloc(spectra*sizeof(float)))) {
            return 0;
         }
      }
   }

   // Allocate memory for covariance matrix determinants
   if(!(trainData->covarianceDeterminants=(float *) malloc(classes*sizeof(float)))) {
      return 0;
   }

   // Allocate memory for priors
   if(!(trainData->priors=(float *) malloc(classes*sizeof(float)))) {
      return 0;
   }

   // Allocate memory for size means
   if(!(trainData->sizeMeans=(float *) malloc(classes*sizeof(float)))) {
      return 0;
   }

   // Allocate memory for size variances
   if(!(trainData->sizeVariances=(float *) malloc(classes*sizeof(float)))) {
      return 0;
   }

   return 1;
}

// Free memory allocated for training data data structure
int freeTrainingDataStruct(struct trainDataStruct* trainData)
{
   int i,j;

   // Free means
   for(i=0;i<trainData->classes;i++) {
      free(trainData->means[i]);
   }
   free(trainData->means);
   trainData->means=NULL;

   // Free inverse covariances matrices
   for(i=0;i<trainData->classes;i++) {
      for(j=0;j<trainData->spectra;j++) {
         free(trainData->covarianceInverses[i][j]);
      }
      free(trainData->covarianceInverses[i]);
   }
   free(trainData->covarianceInverses);
   trainData->covarianceInverses=NULL;

   // Free covariance matrix determinants
   free(trainData->covarianceDeterminants);
   trainData->covarianceDeterminants=NULL;

   // Free priors
   free(trainData->priors);
   trainData->priors=NULL;

   // Free size means
   free(trainData->sizeMeans);
   trainData->sizeMeans=NULL;

   // Free size variances
   free(trainData->sizeVariances);
   trainData->sizeVariances=NULL;

   return 1;
}

// Read training data from file (also allocates memory for data structure)
int readTrainingData(struct trainDataStruct* trainData, char* MFISHdirectory, char* imageName)
{
   // Should be able to determine classes and spectra from file
   const int classes=24;
   const int spectra=5;
   FILE* fileptr;
   char filename[FILENAME_MAX];
   int i,j,k;

   initTrainingDataStruct(trainData, classes, spectra);

   // Read means
   sprintf(filename, "%s%c%s%c%c%c%c%s", MFISHdirectory, FILESEP,
            "train", FILESEP, imageName[0], imageName[1], imageName[2], "m");
   if(!(fileptr=fopen(filename, "r"))) {
      fprintf(stderr, "File '%s' could not be opened", filename);
      exit(1);
   }
   for(i=0;i<classes;i++) {
      for(j=0;j<spectra;j++) {
         fscanf(fileptr, "%f", &(trainData->means[i][j]));
      }
   }
   fclose(fileptr);

   // Read inverse covariances matrices
   sprintf(filename, "%s%c%s%c%c%c%c%s", MFISHdirectory, FILESEP,
            "train", FILESEP, imageName[0], imageName[1], imageName[2], "cinv");
   if(!(fileptr=fopen(filename, "r"))) {
      fprintf(stderr, "File '%s' could not be opened", filename);
      exit(1);
   }
   for(i=0;i<classes;i++) {
      for(j=0;j<spectra;j++) {
         for(k=0;k<spectra;k++) {
            fscanf(fileptr, "%f", &(trainData->covarianceInverses[i][j][k]));
         }
      }
   }
   fclose(fileptr);

   // Read covariance matrix determinants
   sprintf(filename, "%s%c%s%c%c%c%c%s", MFISHdirectory, FILESEP,
            "train", FILESEP, imageName[0], imageName[1], imageName[2], "cdet");
   if(!(fileptr=fopen(filename, "r"))) {
      fprintf(stderr, "File '%s' could not be opened", filename);
      exit(1);
   }
   for(i=0;i<classes;i++) {
      fscanf(fileptr, "%f", &(trainData->covarianceDeterminants[i]));
   }
   fclose(fileptr);

   // Read priors
   sprintf(filename, "%s%c%s%c%c%c%c%s", MFISHdirectory, FILESEP,
            "train", FILESEP, imageName[0], imageName[1], imageName[2], "p");
   if(!(fileptr=fopen(filename, "r"))) {
      fprintf(stderr, "File '%s' could not be opened", filename);
      exit(1);
   }
   for(i=0;i<classes;i++) {
      fscanf(fileptr, "%f", &(trainData->priors[i]));
   }
   fclose(fileptr);

   // Read size means
   sprintf(filename, "%s%c%s%c%s", MFISHdirectory, FILESEP,
            "train", FILESEP, "sizes");
   if(!(fileptr=fopen(filename, "r"))) {
      fprintf(stderr, "File '%s' could not be opened", filename);
      exit(1);
   }
   for(i=0;i<classes;i++) {
      fscanf(fileptr, "%f", &(trainData->sizeMeans[i]));
   }
   fclose(fileptr);

   // Read size variances
   sprintf(filename, "%s%c%s%c%s", MFISHdirectory, FILESEP,
            "train", FILESEP, "variances");
   if(!(fileptr=fopen(filename, "r"))) {
      fprintf(stderr, "File '%s' could not be opened", filename);
      exit(1);
   }
   for(i=0;i<classes;i++) {
      fscanf(fileptr, "%f", &(trainData->sizeVariances[i]));
   }
   fclose(fileptr);

   return 1;
}

// Calculate a particular value of a multi-dimensional Gaussian function
// Derived from Mehul's classification code
float calcGaussian(float* x, float* mean, float** invcov, float covdet, int dimensions)
{
   int i,j;
   double tempVec[dimensions];
   double mvProduct[dimensions];
   double vvProduct=0.0;
   double sum;

   for (i=0;i<dimensions;i++) {
      tempVec[i]=x[i]-mean[i];
   }

   // Matrix-vector multiplication
   for (i=0;i<dimensions;i++) {
      sum=0;
      for (j=0;j<dimensions;j++) {
         sum+=(tempVec[j]*invcov[i][j]);
      }
      mvProduct[i]=sum;
   }

   // Vector-vector multiplication
   for (i=0;i<dimensions;i++) {
      vvProduct+=(mvProduct[i]*tempVec[i]);
   }

   return 1/(sqrt(pow(2*M_PI,dimensions)*covdet*exp(vvProduct)));
}

// Calculate probability of each pixel belonging to each class
int calcProbabilities(MFISH mfImage, unsigned char** mask, struct trainDataStruct* trainData, float*** probs, int rows, int cols)
{
   int i,j,k,l;
   float pixelVector[trainData->spectra];
   double probSum;
   double vectorMag;

   for(i=0;i<rows;i++) {
      for(j=0;j<cols;j++) {
         if(mask[i][j]) {
            probSum=0.0;
            for(k=0;k<trainData->classes;k++) {
               // Normalize vector; ignore DAPI channel
               vectorMag=0.0;
               for(l=0;l<trainData->spectra;l++) {
                  vectorMag+=mfImage[l+1][i][j] * mfImage[l+1][i][j];
               }
               vectorMag=sqrt(vectorMag);
               for(l=0;l<trainData->spectra;l++) {
                  pixelVector[l]=mfImage[l+1][i][j]/vectorMag;
               }

               probs[i][j][k]=calcGaussian(pixelVector,
                                           trainData->means[k],
                                           trainData->covarianceInverses[k],
                                           trainData->covarianceDeterminants[k],
                                           trainData->spectra);
               probSum+=probs[i][j][k];
            }
            for(k=0;k<trainData->classes;k++) {
               probs[i][j][k]/=probSum;
            }
         }
         else {
            for(k=0;k<trainData->classes;k++) {
               probs[i][j][k]=0.0;
            }
         }
      }
   }

   return 1;
}

// Select most likely class for each pixel
int classifyPixels(float*** probs, unsigned char** mask, unsigned char** outputImage, int rows, int cols, int classes)
{
   int i,j,k;
   float maximumProbability;

   for(i=0;i<rows;i++) {
      for(j=0;j<cols;j++) {
         if(mask[i][j]) {
            maximumProbability=0;
            for(k=0;k<classes;k++) {
               if(probs[i][j][k]>maximumProbability) {
                  maximumProbability=probs[i][j][k];
                  outputImage[i][j]=k+1;
               }
            }
         }
         else {
            outputImage[i][j]=0;
         }
      }
   }

   return 1;
}
