/*
Copyright (c) 2001 The University of Texas
All Rights Reserved.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

The GNU Public License is available in the file LICENSE, or you
can write to the Free Software Foundation, Inc., 59 Temple Place -
Suite 330, Boston, MA 02111-1307, USA, or you can find it on the
World Wide Web at http://www.fsf.org.

Programmers:    Wade Schwartzkopf
Version:        @(#)segmentation.c    2.0     08/16/00

The author performed this work while he was with the Department of
Electrical and Computer Engineering, The University of Texas at Austin,
Austin, TX.  Wade Schwartzkopf can be reached at wade@schwartzkopf.org.
The professors supervising the work were Prof. Alan C. Bovik,
bovik@ece.utexas.edu, and Prof. Brian L. Evans, bevans@ece.utexas.edu.
*/

#include <math.h>
#include <stdlib.h>
#include "segmentation.h"
#include "my_matrix_stuff.h"
#include "miscfunc.h"

int drawLine(struct point start, struct point end, unsigned char** outputImage, unsigned char** mask, int lineValue)
{
   int i,dx,dy,length;
   int xIndex, yIndex;
   float x,y,xinc,yinc;

   /* Is cut line contained within object? */
   dx = end.x - start.x;
   dy = end.y - start.y;
   length = maxInt(abs(dx), abs(dy));
   xinc = ((float) dx)/((float) length);
   yinc = ((float) dy)/((float) length);

   // Check if line is valid (if it is contained in mask)
   x = start.x; y = start.y;
   for(i = 0; i <= length; i++) {
      yIndex=(int) rint(y);
      xIndex=(int) rint(x);
      if((mask[yIndex][xIndex])<1) {
         return 0;
      }
      x+=xinc;
      y+=yinc;
   }

   // Draw line
   x = start.x; y = start.y;
   for(i = 0; i <= length; i++) {
      yIndex=(int) rint(y);
      xIndex=(int) rint(x);
      outputImage[yIndex][xIndex]=lineValue;
      x+=xinc;
      y+=yinc;
   }

   return 1;
}

int joinOverlappedEnds(unsigned char** subsegments, unsigned char** mask, int lowX, int highX, int lowY, int highY)
{
   unsigned char** temp=allocateByteMatrix(highY+1, highX+1);
   struct pointList* boundaryPoints=NULL;
   struct pointList* tempptr_i;
   struct pointList* tempptr_j;
   int segmentsConnected=0;

   // Find boundary of segments and possibly overlapping object
   findBoundary(subsegments, mask, temp, NULL, lowX, highX, lowY, highY);
   boundaryPoints=listPoints(temp, lowX, highX, lowY, highY);
   if(boundaryPoints) {
      // Connect subsegments
      for(tempptr_i = boundaryPoints; tempptr_i->next; tempptr_i = tempptr_i->next) {
         for(tempptr_j = tempptr_i->next; tempptr_j; tempptr_j = tempptr_j->next) {
            segmentsConnected|=drawLine(tempptr_i->value, tempptr_j->value, subsegments, mask, 255);
         }
      }
   }
   // Fill any remaining holes
   findHoles(subsegments, temp, lowX, highX, lowY, highY);
   orImage(subsegments, temp, subsegments, lowX, highX, lowY, highY);

   // Have all segments been connected
   if(segmentsConnected) {
      if(labelObjects(subsegments, subsegments, temp, 1, lowX, highX, lowY, highY, 4)>1) {
         segmentsConnected=0;
      }
   }

   freeByteMatrix(temp);
   freePointList(boundaryPoints);

   return segmentsConnected;
}

struct labelList* allClasses(int classes)
{
   int i;
   struct labelList* currentLabel=NULL;

   for(i=0;i<classes;i++) {
      currentLabel=addLabel(currentLabel, i);
   }

   return currentLabel;
}

double calcObjProb(MFISH inputMFISH, float*** probs, unsigned char** mask, unsigned char** overlapMask, struct trainDataStruct* trainData, struct labelList* possibleLabels, int totalArea, int useSize, int lowX, int highX, int lowY, int highY, int* mlClass)
{
   int i,j,k;
   int shouldFreeLabels=0;
   double prejoinedArea, sizeConfidence;
   double objArea=0;
   double sizeDiff;
   /* double vectorMag, probSum; */
   double maxProb;
   double* avgProbs=malloc(trainData->classes*sizeof(double));
   double* sizeProbs=malloc(trainData->classes*sizeof(double));
   double* totalProbs=malloc(trainData->classes*sizeof(double));
   float* avgPixel=malloc(trainData->spectra*sizeof(float));
   struct labelList* iptr;
   unsigned char** subsegments=allocateByteMatrix(highY+1, highX+1);
   unsigned char** newMask;

   if(!possibleLabels) {
      possibleLabels=allClasses(trainData->classes);
      shouldFreeLabels=1;
   }

   // If more than one connected component (overlapped object), join segments
   prejoinedArea=findArea(mask, lowX, highX, lowY, highY);
   if(labelObjects(mask, mask, subsegments, 1, lowX, highX, lowY, highY, 4)>1) {
      if(!joinOverlappedEnds(subsegments, overlapMask, lowX, highX, lowY, highY)) {
         // No valid connection found between the two subsegments
         if(shouldFreeLabels) { freeLabelList(possibleLabels); }
         freeByteMatrix(subsegments);
         free(totalProbs);
         free(sizeProbs);
         free(avgProbs);
         free(avgPixel);
         return 0.0;
      }
      newMask=subsegments;
   }
   else {
      newMask=mask;
   }

   // Calculate average pixel probabilities for each class
   for(i=0;i<trainData->spectra;i++) { avgPixel[i]=0.0; }
   for(iptr=possibleLabels; iptr; iptr=iptr->next)
      { avgProbs[iptr->value]=0.0; }

   for(i = lowY; i <= highY; i++) {
      for(j = lowX; j <= highX; j++) {
         if (newMask[i][j]) {
            objArea++;
            for(iptr=possibleLabels; iptr; iptr=iptr->next) {
               avgProbs[iptr->value]+=probs[i][j][iptr->value];
            }
            for(k=0;k<trainData->spectra;k++) {
               avgPixel[k]+=inputMFISH[k+1][i][j];
            }
         }
      }
   }
   for(i=0;i<trainData->spectra;i++) {
      avgPixel[i]/=objArea;
   }
   /* probSum=0; */
   for(iptr=possibleLabels; iptr; iptr=iptr->next) {
      avgProbs[iptr->value]/=objArea; 
      // Normalize vector; ignore DAPI channel
      /* vectorMag=0.0;
      for(k=0;k<trainData->spectra;k++) {
         vectorMag+=avgPixel[k] * avgPixel[k];
      }
      vectorMag=sqrt(vectorMag);
      for(k=0;k<trainData->spectra;k++) {
         avgPixel[k]/=vectorMag;
      }
      avgProbs[iptr->value]=calcGaussian(avgPixel,
                               trainData->means[iptr->value],
                               trainData->covarianceInverses[iptr->value],
                               trainData->covarianceDeterminants[iptr->value],
                               trainData->spectra);
      probSum+=avgProbs[iptr->value]; */
   }
   /* for(iptr=possibleLabels; iptr; iptr=iptr->next) {
      avgProbs[iptr->value]/=probSum;
   } */

   // Calculate object size probabilities for each class
   for(iptr=possibleLabels; iptr; iptr=iptr->next) {
      if(totalArea) {
         sizeDiff=trainData->sizeMeans[iptr->value]-(objArea/(double)totalArea);
         sizeProbs[iptr->value]=exp(-sizeDiff*sizeDiff/(2*trainData->sizeVariances[iptr->value]));
      }
      else {
         sizeProbs[iptr->value]=1;
      }
   }
   sizeConfidence=prejoinedArea/objArea;  // Percentage of object showing

   // Calculate total probabilities for each class
   maxProb=0;
   for(iptr=possibleLabels; iptr; iptr=iptr->next) {
      if(useSize) {
         totalProbs[iptr->value]=avgProbs[iptr->value]*sizeProbs[iptr->value]*sizeConfidence;
      }
      else {
         totalProbs[iptr->value]=avgProbs[iptr->value];
      }
      if(totalProbs[iptr->value]>maxProb) {
         maxProb=totalProbs[iptr->value];
         if(mlClass) { *mlClass=iptr->value+1; }
      }
   }

   if(shouldFreeLabels) { freeLabelList(possibleLabels); }
   freeByteMatrix(subsegments);
   free(totalProbs);
   free(sizeProbs);
   free(avgProbs);
   free(avgPixel);

   return maxProb;
}

int reclassify(MFISH MFISHimage, float*** probs, unsigned char** inputImage, unsigned char** mask, unsigned char** outputImage, struct trainDataStruct* trainData, int totalArea, int lowX, int highX, int lowY, int highY)
{
   int i,j;
   int mlClass;
   struct labelList* boundaryValues=NULL;
   unsigned char** tempImage=allocateByteMatrix(highY+1, highX+1);
   unsigned char** single=allocateByteMatrix(highY+1, highX+1);

   for(i=lowY;i<=highY;i++) {
      for(j=lowX;j<highX;j++) {
         if(mask[i][j]&&(outputImage[i][j]<1)) {
            mlClass=inputImage[i][j];
            clearImage(single, lowX, highX, lowY, highY);
            if(THRESH_SIZE>labelSingleObject(inputImage, mask, single, j, i,
                                             1, 0, lowX, highX,
                                             lowY, highY, 4)) {
               findBoundary(single, inputImage, tempImage, &boundaryValues,
                            lowX, highX, lowY, highY);
               if(boundaryValues) {
                  if(boundaryValues->next) {
                     calcObjProb(MFISHimage, probs, single, inputImage, trainData, boundaryValues, 0, 1, lowX, highX, lowY, highY, &mlClass);
                  }
                  else {
                     mlClass=boundaryValues->value+1;
                  }
                  freeLabelList(boundaryValues);
                  boundaryValues=NULL;
               }
            }
            labelSingleObject(inputImage, mask, outputImage, j, i, mlClass, 0, lowX, highX, lowY, highY, 4);
         }
      }
   }

   freeByteMatrix(tempImage);
   freeByteMatrix(single);

   return 1;
}


int isCluster(MFISH inputMFISH, float*** probs, unsigned char** mask, struct trainDataStruct* trainData, int totalArea, int lowX, int highX, int lowY, int highY, int* mlClass, double* likelihood)
{
   return ((*likelihood=calcObjProb(inputMFISH, probs, mask, mask, trainData, NULL, totalArea, 1, lowX, highX, lowY, highY, mlClass))<THRESH_PROB);
}

int betterMerge(double prob1, double prob2, double jointProb, double* oldMeasure)
{
   // if((jointProb>=((prob1+prob2)/2))&&(jointProb>(*oldMeasure))) {
   if((jointProb>=sqrt(prob1*prob2))&&(jointProb>(*oldMeasure))) {
      *oldMeasure=jointProb;
      return 1;
   }

   return 0;
}

int rejoin(MFISH inputMFISH, float*** probs, unsigned char** subsegments, struct trainDataStruct* trainData, int totalArea, int lowX, int highX, int lowY, int highY)
{
   // Can be made much more efficient if, after the first iteration of
   //    the while loop, it only recalculates newly created segment
   int merged=1;
   int i,j;
   int label1=0, label2=0;
   int numSubsegments;
   double prob1, prob2, jointProb;
   double bestMergeMeasure;
   unsigned char** temp=allocateByteMatrix(highY+1, highX+1);
   
   numSubsegments=max(subsegments, lowX, highX, lowY, highY);
   while(merged) {
      merged=0;
      bestMergeMeasure=0.0;
      for(i=1; i<numSubsegments; i++) {
         find_equal(subsegments, temp, i, lowX, highX, lowY, highY);
         prob1=calcObjProb(inputMFISH, probs, temp, subsegments, trainData, NULL, totalArea, 1, lowX, highX, lowY, highY, NULL);
         for(j=i+1; j<=numSubsegments; j++) {
            printf("      Object %d prob = %1.2f | ", i, prob1);
            find_equal(subsegments, temp, j, lowX, highX, lowY, highY);
            prob2=calcObjProb(inputMFISH, probs, temp, subsegments, trainData, NULL, totalArea, 1, lowX, highX, lowY, highY, NULL);
            printf("Object %d prob = %1.2f | ", j, prob2);
            find_two_equal(subsegments, temp, i, j, lowX, highX, lowY, highY);
            jointProb=calcObjProb(inputMFISH, probs, temp, subsegments, trainData, NULL, totalArea, 1, lowX, highX, lowY, highY, NULL);
            printf("Joined prob = %1.4f\r\n", jointProb);
            if(betterMerge(prob1, prob2, jointProb, &bestMergeMeasure)) {
               merged=1;
               label1=i;
               label2=j;
            }
         }
      }
      if(merged) {
         printf("   Joining objects %d and %d...\r\n", label1, label2);
         relabel(subsegments, subsegments, label2, label1, lowX, highX, lowY, highY);
         skiplabel(subsegments, label2, lowX, highX, lowY, highY);
         numSubsegments--;
      }
   }

   freeByteMatrix(temp);

   return numSubsegments;
}


void segment(MFISH inputMFISH, unsigned char** mask, struct trainDataStruct* trainData, unsigned char** outputImage, int rows, int cols)
{
   FILE* likelihoods;
   int numSegments, numSubsegments;
   int totalPixels;
   int i,j,ssCounter;
   int lowX, highX, lowY, highY;
   int mlClass, mlClassNoSize;
   double objProb;
   unsigned char** classifiedPixels=allocateByteMatrix(rows, cols);
   unsigned char** majFilt=allocateByteMatrix(rows, cols);
   unsigned char** labels=allocateByteMatrix(rows, cols);
   unsigned char** single=allocateByteMatrix(rows, cols);
   unsigned char** subsegments=allocateByteMatrix(rows, cols);
   unsigned char** reclass=allocateByteMatrix(rows, cols);
   unsigned char** noSize=allocateByteMatrix(rows, cols);
   unsigned char** flagged=allocateByteMatrix(rows, cols);
   unsigned char** segLabels=allocateByteMatrix(rows, cols);
   float*** probs=allocate3DFloatMatrix(rows, cols, trainData->classes);

   // Calculate probabilites of each pixel belonging to each class
   printf("Calculating probabilities...\r\n");
   calcProbabilities(inputMFISH, mask, trainData, probs, rows, cols);

   // Select most likely class
   printf("Classifying pixels...\r\n");
   classifyPixels(probs, mask, classifiedPixels, rows, cols, trainData->classes);
   writeImageAs2DByte("images/pixels.png", classifiedPixels, rows, cols, 3);

   // Perform majority filter
   // (not necessary here for algorithm, but good to see what's going on)
   printf("Performing majority filter...\r\n");
   clearImage(majFilt, 0, cols-1, 0, rows-1);
   majorityFilter(classifiedPixels, mask, majFilt, 8, 0, cols-1, 0, rows-1);
   writeImageAs2DByte("images/majfilt.png", majFilt, rows, cols, 3);

   // Label connected components
   numSegments= labelObjects(mask, mask, labels, 1, 0, cols-1, 0, rows-1, 4);
   writeImageAs2DByte("images/labels.png", labels, rows, cols, 3);

   // Calculate total chromosome area
   totalPixels=findArea(mask, 0, cols-1, 0, rows-1);

   // Remove small segments
   // (not necessary here for algorithm, but good to see what's going on)
   printf("Reclassifying small segments...\r\n");
   clearImage(reclass, 0, cols-1, 0, rows-1);
   reclassify(inputMFISH, probs, majFilt, majFilt, reclass, trainData, totalPixels, 0, cols-1, 0, rows-1);
   writeImageAs2DByte("images/reclass.png", reclass, rows, cols, 3);

   clearImage(outputImage, 0, cols-1, 0, rows-1);
   clearImage(noSize, 0, cols-1, 0, rows-1);
   clearImage(segLabels, 0, cols-1, 0, rows-1);
   ssCounter=1;
   likelihoods=fopen("images/likelihoods.txt", "w");
   // Look for unlikely objects
   for(i = 1; ((i <= numSegments)&&(i<256)); i++) {
      printf("Checking object %d of %d...\r\n", i, numSegments);
      find_equal(labels, single, i, 0, cols-1, 0, rows-1);
      boundingBox(single, &lowX, &highX, &lowY, &highY, rows, cols);
      if(isCluster(inputMFISH, probs, single, trainData, totalPixels, lowX, highX, lowY, highY, &mlClass, &objProb)) {
         printf("   Object %d low probability.  Examining further.\r\n", i);
         // Split object using majority filtering of classified pixels
         clearImage(majFilt, lowX, highX, lowY, highY);
	 clearImage(reclass, lowX, highX, lowY, highY);
         majorityFilter(classifiedPixels, single, majFilt, 8, lowX, highX, lowY, highY);
         reclassify(inputMFISH, probs, majFilt, majFilt, reclass, trainData, totalPixels, lowX, highX, lowY, highY);
         // Then recombine segments to get reasonable chromosomes
         numSubsegments=labelObjects(reclass, single, subsegments, 0, lowX, highX, lowY, highY, 4);
         numSubsegments=rejoin(inputMFISH, probs, subsegments, trainData, totalPixels, lowX, highX, lowY, highY);
         // Classify new segments as most likely class
         for(j=1;j<=numSubsegments;j++) {
            clearImage(single, lowX, highX, lowY, highY);
            find_equal(subsegments, single, j, lowX, highX, lowY, highY);
            objProb=calcObjProb(inputMFISH, probs, single, subsegments, trainData, NULL, totalPixels, 1, lowX, highX, lowY, highY, &mlClass);
            fprintf(likelihoods,"%1.2f\r\n",objProb);
            calcObjProb(inputMFISH, probs, single, subsegments, trainData, NULL, totalPixels, 0, lowX, highX, lowY, highY, &mlClassNoSize);
            relabel(subsegments, outputImage, j, mlClass, lowX, highX, lowY, highY);
            relabel(subsegments, noSize, j, mlClassNoSize, lowX, highX, lowY, highY);
            relabel(subsegments, segLabels, j, ssCounter++, lowX, highX, lowY, highY);
            if(objProb<THRESH_PROB) {
               relabel(subsegments, flagged, j, 1, lowX, highX, lowY, highY);
            }
         }
      }
      else {
         // Label object as most likely class
         relabel(labels, outputImage, i, mlClass, lowX, highX, lowY, highY);
         find_equal(labels, single, i, lowX, highX, lowY, highY);
         fprintf(likelihoods,"%1.2f\r\n",objProb);
         calcObjProb(inputMFISH, probs, single, single, trainData, NULL, totalPixels, 0, lowX, highX, lowY, highY, &mlClassNoSize);
         relabel(labels, noSize, i, mlClassNoSize, lowX, highX, lowY, highY);
         relabel(labels, segLabels, i, ssCounter++, lowX, highX, lowY, highY);
      }
   }

   writeImageAs2DByte("images/nosize.png", noSize, rows, cols, 3);
   writeImageAs2DByte("images/flag.png", flagged, rows, cols, 3);
   writeImageAs2DByte("images/segLabels.png", segLabels, rows, cols, 3);
   fclose(likelihoods);

   freeByteMatrix(classifiedPixels);
   freeByteMatrix(majFilt);
   freeByteMatrix(labels);
   freeByteMatrix(single);
   freeByteMatrix(subsegments);
   freeByteMatrix(reclass);
   freeByteMatrix(noSize);
   freeByteMatrix(flagged);
   freeByteMatrix(segLabels);
   free3DFloatMatrix(probs, rows, cols);
}
